/*
**  Copyright (c) 1991 Microsoft Corporation
*/
//===========================================================================
// FILE                         qppm.h
//
// MODULE                       Print Manager / Queue Processor interface
//
// PURPOSE                      This file declares the interface between
//                              the Windows Print Manager and QPs
//
//
//===========================================================================


// QUEUE PROCESSOR interface
//
// These are the declarations for the services provided by the queue
// processor (QP) to the Print Manager (PM).
//
//
// QP_INIT - Initialize the queue processor.
//          
// wParam - Version number of interface expected, should equal PMQP_VERS.
// lParam - Far pointer to QPINITDATA struct.
// return - The function returns a non zero handle to its internal port
//          data structure if successful, or zero if there was an error or
//          the specified printer/port combination is not supported.
//          If zero is returned, wErr in the QPINITDATA structure
//          will be set by the QP.
//
//
// QP_BEGINJOB - Start a new print job.
//
// wParam - The handle to the port data returned from QP_INIT.
// lParam - Far pointer to QPBEGINDATA struct.
// return - Returns a non-zero handle to the new job if successful, or
//          else NULL, in which case, uiErr of the QPBEGINDATA structure
//          will be set to one of the PM_PORT_* values.
//
//
// QP_WRITE - Write data to the queue processor.
//
// wParam - The handle to the port data returned from QP_INIT.
// lParam - Far pointer to QPWRITEDATA struct.
// return - The function returns the number of bytes consumed
//          and fills in the wErr word of the QPWRITEDATA struct with
//          the PM_PORT_* device status constant.
//
//
// QP_ENDJOB - End the print job, the PM has no more data for this job.
//
// wParam - The handle to the port data returned from QP_INIT.
// lParam - Far pointer to QPENDJOB struct.
// return - 0 if the job was not closed, 1 if it was.
//          NOTE: if the QP is just waiting for a page to clear the printer
//          or something, it would be smart to return 1 and start receiving
//          data for the next job. If there is no next job, the QP will
//          get a chance to fail on the QP_CLOSE call.
//
//
// QP_DISPLAY - Instructs the queue processor to display itself, if it is not
//          already, and to make itself the active window.  The user has
//          pushed the Status button on the toolbar or chosen Printer Status
//          from the View menu.
//
// wParam - The handle to the port data returned from QP_INIT.
// lParam - Far pointer to QPDISPLAYDATA struct.
// return - Zero.
//
//
// QP_IDLE - Gives the QP a chance to run when there are no jobs
//          in the queue.
//
// wParam - The handle to the port data returned from QP_INIT.
// lParam - Far pointer to QPIDLEDATA struct.
// lParam - Zero.
// return - Zero.
//
//
// QP_CLOSE - End communications with the queue processor.
//
// wParam - The handle to the port data returned from QP_INIT, which
//          becomes invalid if this call succeeds.
// lParam - Far pointer to QPCLOSEDATA struct.
// return - The function returns 1 if it is okay to close, or zero
//          if the queue processor is still working.
//
//
// QP_GETSPOOL - Retain spool file when spooling and collating copies
//          or manual duplexing to later resend file(s) to printer.
//
// wParam - The handle to the port data returned from QP_INIT.
// lParam - Far pointer to SUBSTORE struct; a subset of STORE struct
// return - TRUE if file successfully retained. FALSE if error.

// QPAction - The action routine the print manager calls, its behavior
// depends upon the action value, as described above.

#define QPEXPORT _loadds _far _pascal

extern UINT QPEXPORT QPAction(UINT uiAction, UINT uiParam, LPVOID lParam);

typedef UINT (QPEXPORT FAR *LPQPACTION)(UINT uiAction, UINT uiParam, LPVOID lParam);

#define fn_QPAction(lpfnQPAction, uiAction, uiParam, lParam) \
                    ((*(LPQPACTION)(lpfnQPAction))((uiAction),(uiParam),(lParam)) )


// ERROR VALUES
//
// Error constants

#define PM_PORT_READY       0
#define PM_PORT_GENERROR    1
#define PM_PORT_BUSY        2
#define PM_PORT_PAUSED      3


// PMQP_VERS
//
// Version number of print manager/queue processor interface.

#define PMQP_VERS           0x0026


// Ordinal for loading proc address from QP dll.

#define ID_FN_QPACTION      100


// Action numbers.

#define QP_INIT             1
#define QP_BEGINJOB         2
#define QP_WRITE            3
#define QP_ENDJOB           4
#define QP_DISPLAY          5
#define QP_IDLE             6
#define QP_CLOSE            7
#define QP_GETSPOOL         8
#define QP_QUERYCLOSE       9


// Parameter data for individual action routines.

typedef struct _QPHDR           // Common part of all following structures
{
    HWND        hwndPM;         // window QP should post to and use as parent
    UINT        uiErr;          // If function fails, shows reason
    int         iPriority;      // spooler background priority (0=hi 2=low)
                                // PRIVATE: if set to -1, caller is driver direct
    DWORD       ulIdleRate;     // how often QP wants to be called when idle
    UINT        uiReserved;     // reserved
} QPHDR, *PQPHDR, FAR *LPQPHDR;


typedef struct _QPINITDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;     // PRIVATE: if QP sets to 0xFFFF, PM doesn't
                                // call Write (BeginJob and EndJob only)
    LPSTR       lpszDriver;     // far pointer to printer driver name (JUMBO)
    LPSTR       lpszPort;       // far pointer to port name (LPT2:)
    LPSTR       lpszPrinter;    // far pointer to printer name (JUMBO for LaserJet II)
} QPINITDATA, *PQPINITDATA, FAR *LPQPINITDATA;


typedef struct _QPBEGINDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;     // PRIVATE: if caller is driver, this is HDC
    LPSTR       lpszJobName;    // far pointer to job name
//    DWORD       dwPreview;      // first 4 bytes of spool data
} QPBEGINDATA, *PQPBEGINDATA, FAR *LPQPBEGINDATA;


typedef struct _QPWRITEDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;
    HANDLE      hJob;           // job handle returned from BeginJob
    LPVOID      lpData;         // far pointer to buffer
    UINT        uiCount;        // count of bytespointed to by lpData
} QPWRITEDATA, *PQPWRITEDATA, FAR *LPQPWRITEDATA;


typedef struct _QPDISPLAYDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;
} QPDISPLAYDATA, *PQPDISPLAYDATA, FAR *LPQPDISPLAYDATA;


typedef struct _QPIDLEDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;
} QPIDLEDATA, *PQPIDLEDATA, FAR *LPQPIDLEDATA;


typedef struct _QPENDJOBDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;
    HANDLE      hJob;           // job handle returned from BeginJob
    BOOL        bAbort;         // job is to be aborted, must return 1
} QPENDJOBDATA, *PQPENDJOBDATA, FAR *LPQPENDJOBDATA;


typedef struct _QPCLOSEDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;
    BOOL        bMustClose;     // If TRUE, QP not allowed to return failure.
} QPCLOSEDATA, *PQPCLOSEDATA, FAR *LPQPCLOSEDATA;


typedef struct _QPQUERYCLOSEDATA
{
    HWND        hwndPM;
    UINT        uiErr;
    int         iPriority;
    DWORD       ulIdleRate;
    UINT        uiReserved;
} QPQUERYCLOSEDATA, *PQPQUERTCLOSEDATA, FAR *LPQPQUERYCLOSEDATA;


// Constants for uiFlags and uiErr

// uiErr in QPINITDATA
#define QP_INITERR_WRONGVERSION     1   // wParam doesn't match PMQP_VERS
#define QP_INITERR_DRVNOTSUPPORTED  2   // QP doesn't support printer driver/port
#define QP_INITERR_INITFAILED       3   // general initialization failure
#define QP_INITERR_PORTNOTSUPPORTED 4   // doesn't support port or redirection
#define QP_INITERR_NOMEM            5   // out of memory
