//==========================================================================;
//
//  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
//  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
//  PURPOSE.
//
//  Copyright (c) 1994  Microsoft Corporation.  All Rights Reserved.
//
//--------------------------------------------------------------------------;
//
//  profile.c
//
//  Description:
//
//      This file contains routines which access the Registry (for Win32)
//      using the same interface as the Win16 Profile APIs.  This allows
//      the use of common source code for Win16/Win32 cross-compilation.
//      This code reads profiles from the specified .ini file under Win16,
//      and read from
//
//          HKEY_LOCAL_MACHINE\Software\Microsoft\Multimedia\<subkey>
//
//      for Win32.  Note that the Private routines ignore the filename for
//      Win32.
//
//      The source code must be changed in one respect.  In Win16 there is
//      no WriteProfileInt, so if you wanted to write an integer you would
//      probably write the integer to a string using wsprintf, then call
//      WriteProfileString to write it.  Then you would read it using
//      GetProfileInt.  This sequence will still work in Win16, but Win32
//      saves the type of the data in registry, as well as the data itself.
//      So, it will try to read a REG_DWORD from a REG_SZ entry and will
//      fail.  Therefore, if you want to read using mmGetProfileInt, you
//      must write using mmWriteProfileInt.  For Win32, this writes a DWORD
//      to the registry; for Win16, this writes the integer to a string and
//      then calls WriteProfileString.  There is also a
//      mmWritePrivateProfileInt.
//
//      You also might want to change the source code to reflect the fact
//      that the new APIs use DWORDs and UINTs rather than ints.  
//
//      Here are the routines:
//
//          mmGetProfileInt
//          mmGetPrivateProfileInt
//          mmGetProfileString
//          mmGetPrivateProfileString
//          mmWriteProfileInt
//          mmWritePrivateProfileInt
//          mmWriteProfileString
//          mmWritePrivateProfileString
//
//==========================================================================;

#include <windows.h>
#include <windowsx.h>

#include "debug.h"
#include "profile.h"


#ifdef WIN32

//==========================================================================;
//
//      WIN 32
//
//==========================================================================;


#define PROFILE_DEFAULTKEY  HKEY_CURRENT_USER

const TCHAR gszMultimediaKey[] = TEXT("Software\\Microsoft\\Multimedia\\");


//--------------------------------------------------------------------------;
//  
//  HKEY mmprofOpenKey
//  
//  Description:
//      This routine opens a sub key under the default MM key.
//  
//  Arguments:
//      LPCTSTR lpszKeyName:  Name of the sub key.
//      DWORD   dwAccess:     Access writes desired:  KEY_READ or KEY_WRITE.
//  
//  Return (HKEY):  Handle to the opened key, or NULL if the request failed.
//  
//--------------------------------------------------------------------------;

HKEY mmprofOpenKey
(
    LPCTSTR lpszKeyName,
    DWORD   dwAccess
)
{
    HKEY    hkeyMM = NULL;
    HKEY    hkeyRet = NULL;


    RegCreateKeyEx( PROFILE_DEFAULTKEY, gszMultimediaKey, 0, NULL, 0,
                       KEY_WRITE, NULL, &hkeyMM, NULL );

    if( NULL != hkeyMM )
    {
        if( KEY_WRITE == dwAccess )
        {
            RegCreateKeyEx( hkeyMM, lpszKeyName, 0, NULL, 0,
                        KEY_WRITE, NULL, &hkeyRet, NULL );
        }
        else
        {
            RegOpenKeyEx( hkeyMM, lpszKeyName, 0, KEY_READ, &hkeyRet );
        }

        RegCloseKey( hkeyMM );
    }

    return hkeyRet;
}



//--------------------------------------------------------------------------;
//  
//  HKEY mmprofCloseKey
//  
//  Description:
//      This routine closes a key opened by mmprofOpenKey.
//  
//  Arguments:
//      HKEY hkey:  The key returned by mmprofOpenKey.
//  
//--------------------------------------------------------------------------;

INLINE void mmprofCloseKey( HKEY hkey )
{
    (void)RegCloseKey( hkey );
}



//--------------------------------------------------------------------------;
//  
//  DWORD mmprofStrncpy
//  
//  Description:
//      This routine is an internal copy of strncpy() which handles both
//      Ansi and Unicode characters.
//  
//  Arguments:
//      LPTSTR  lpszDest:
//      LPCTSTR lpszSrc:
//      DWORD   cchDest:     Size of lpszDest (in characters, not bytes!)
//  
//  Return (DWORD):  The number of bytes copied (not including the
//                  terminating '\0'.
//
//--------------------------------------------------------------------------;

DWORD mmprofStrncpy
(
    LPTSTR  lpszDest,
    LPCTSTR lpszSrc,
    DWORD   cchDest
)
{
    DWORD i;

    for( i=0; i<cchDest && *lpszSrc!=TEXT('\0'); i++ )
    {
        *lpszDest++ = *lpszSrc++;
    }

    if( i<cchDest )
    {
        *lpszDest = TEXT('\0');
    }

    return i;   
}



//--------------------------------------------------------------------------;
//  
//  UINT mmGetProfileInt
//  
//  Description:
//      This routine mimics GetProfileInt using the registry.
//  
//  Arguments:
//      LPCTSTR lpszSection:
//      LPCTSTR lpszEntry:
//      UINT    uDefault:
//  
//  Return (UINT):
//  
//--------------------------------------------------------------------------;

UINT mmGetProfileInt
(
    LPCTSTR lpszSection,
    LPCTSTR lpszEntry,
    UINT    uDefault
)
{
    HKEY    hkey;
    DWORD   dwType = (DWORD)~REG_DWORD;  // Init to anything but REG_DWORD.
    DWORD   cbSize = sizeof(DWORD);
    DWORD   dwRet;

    hkey = mmprofOpenKey( lpszSection, KEY_READ );
    if( NULL == hkey )
    {
        return uDefault;
    }

    RegQueryValueEx( hkey, (LPTSTR)lpszEntry, NULL, &dwType, (LPBYTE)&dwRet,
                                                                &cbSize );

    mmprofCloseKey( hkey );

    return (REG_DWORD==dwType) ? (UINT)dwRet : uDefault;
}



//--------------------------------------------------------------------------;
//  
//  DWORD mmGetProfileString
//  
//  Description:
//      This routine mimics GetProfileString using the registry.
//  
//  Arguments:
//      LPCTSTR lpszSection:
//      LPCTSTR lpszEntry:
//      LPCTSTR lpszDefault:
//      LPCTSTR lpszReturnBuffer:
//      DWORD   cchReturnBuffer:     In characters, not bytes!!!
//  
//  Return (DWORD):  Number of characters copied to lpszReturnBuffer, not
//                  including the terminating '\0'.
//  
//--------------------------------------------------------------------------;

DWORD mmGetProfileString
(
    LPCTSTR lpszSection,
    LPCTSTR lpszEntry,
    LPCTSTR lpszDefault,
    LPTSTR  lpszReturnBuffer,
    DWORD   cchReturnBuffer
)
{
    HKEY    hkey;
    DWORD   dwType;
    DWORD   cbSize;
    int     iRet = 0;

    hkey = mmprofOpenKey( lpszSection, KEY_READ );
    if( NULL == hkey )
    {
        return mmprofStrncpy( lpszReturnBuffer, lpszDefault, cchReturnBuffer );
    }

    cbSize = sizeof(TCHAR)*cchReturnBuffer;
    RegQueryValueEx( hkey, (LPTSTR)lpszEntry, NULL, &dwType,
                        (LPBYTE)&lpszReturnBuffer, &cbSize );

    mmprofCloseKey( hkey );

    if( REG_SZ == dwType )
    {
        return cbSize/sizeof(TCHAR) - 1;
    }
    else
    {
        return mmprofStrncpy( lpszReturnBuffer, lpszDefault, cchReturnBuffer );
    }
}

    

//--------------------------------------------------------------------------;
//  
//  BOOL mmWriteProfileInt
//  
//  Description:
//      This routine writes a UINT to the multimedia section of the registry.
//  
//  Arguments:
//      LPCTSTR lpszSection:
//      LPCTSTR lpszEntry:
//      UINT    uOut:       The value to be written.
//  
//  Return (BOOL):  Nonzero if the function was successful.
//  
//--------------------------------------------------------------------------;

BOOL mmWriteProfileInt
(
    LPCTSTR lpszSection,
    LPCTSTR lpszEntry,
    UINT    uOut
)
{
    HKEY    hkey;
    LONG    lRet;
    DWORD   dw;

    hkey = mmprofOpenKey( lpszSection, KEY_WRITE );
    if( NULL == hkey )
    {
        return FALSE;
    }

    dw   = (DWORD)uOut;
    lRet = RegSetValueEx( hkey, lpszEntry, 0, REG_DWORD, (LPBYTE)&dw,
                    sizeof(DWORD) );

    mmprofCloseKey( hkey );

    return (ERROR_SUCCESS==lRet) ? TRUE : FALSE;
}



//--------------------------------------------------------------------------;
//  
//  BOOL mmWriteProfileString
//  
//  Description:
//      This routine mimics WriteProfileString using the registry.
//  
//  Arguments:
//      LPCTSTR lpszSection:
//      LPCTSTR lpszEntry:
//      LPCTSTR lpszString:
//  
//  Return (BOOL):  Nonzero if the function was successful.
//  
//--------------------------------------------------------------------------;

BOOL mmWriteProfileString
(
    LPCTSTR lpszSection,
    LPCTSTR lpszEntry,
    LPCTSTR lpszString
)
{
    HKEY    hkey;
    LONG    lRet;

    hkey = mmprofOpenKey( lpszSection, KEY_WRITE );
    if( NULL == hkey )
    {
        return FALSE;
    }

    lRet = RegSetValueEx( hkey, lpszEntry, 0, REG_SZ, (LPBYTE)lpszString,
                    (DWORD)( sizeof(TCHAR) * (1+lstrlen(lpszString)) )  );

    mmprofCloseKey( hkey );

    return (ERROR_SUCCESS==lRet) ? TRUE : FALSE;
}


#else

//==========================================================================;
//
//      WIN 16
//
//==========================================================================;

const TCHAR gszNumberFormat[] = TEXT("%u");


//--------------------------------------------------------------------------;
//  
//  BOOL mmWriteProfileInt
//  
//  Description:
//      This routine writes a UINT to the specified section of win.ini.
//  
//  Arguments:
//      LPCSTR  lpszSection:
//      LPCSTR  lpszEntry:
//      UINT    uOut:       The value to be written.
//  
//  Return (BOOL):  Nonzero if the function was successful.
//  
//--------------------------------------------------------------------------;

BOOL mmWriteProfileInt
(
    LPCSTR  lpszSection,
    LPCSTR  lpszEntry,
    UINT    uOut
)
{
    TCHAR szOutput[20];

    wsprintf( szOutput, gszNumberFormat, uOut );
    return WriteProfileString( lpszSection, lpszEntry, szOutput );
}



//--------------------------------------------------------------------------;
//  
//  BOOL mmWritePrivateProfileInt
//  
//  Description:
//      This routine writes a UINT to the specified section of the specified
//      .ini file.
//  
//  Arguments:
//      LPCSTR  lpszSection:
//      LPCSTR  lpszEntry:
//      UINT    uOut:          The value to be written.
//      LPCSTR  lpszFilename:  The .ini file name.
//  
//  Return (BOOL):  Nonzero if the function was successful.
//  
//--------------------------------------------------------------------------;

BOOL mmWritePrivateProfileInt
(
    LPCSTR  lpszSection,
    LPCSTR  lpszEntry,
    UINT    uOut,
    LPCSTR  lpszFilename
)
{
    TCHAR szOutput[20];

    wsprintf( szOutput, gszNumberFormat, uOut );
    return WritePrivateProfileString( lpszSection, lpszEntry, szOutput,
                                                            lpszFilename );
}


#endif // WIN32

