//****************************************************************************
// File:    dlgs.c
//
// Purpose: Mailslot dialog procs
//
// Functions:
//    MakeDlgProc()  - processes messages for "DosMakeMailslot" dialog box
//    WriteDlgProc() - processes messages for "DosWriteMailslot" dialog box
//    ErrorMB()      - displays a message box depending on the given error code
//
//****************************************************************************

#include <windows.h>
#include <string.h>
#include <netcons.h>    // required for API_FUNCTION #def
#include <mailslot.h>   // mailslot API prototypes
#include <neterr.h>     // required for NERR_* errors
#define INCL_DOSERRORS
#include <bseerr.h>
#include "global.h"
#include "dlgs.h"

EXTERN UINT hMailslot;    // handle to mailslot

//****************************************************************************
// Function: MakeDlgProc()
//
// Purpose:  Processes messages for "DosMakeMailslot" dialog box
//
// Parameters:
//    HWND   hDlg
//    UINT   message
//    WPARAM wParam
//    LPARAM lParam
//
// Returns:
//    BOOL
//
//****************************************************************************
BOOL FAR PASCAL MakeDlgProc (HWND   hDlg,
                             UINT   message,
                             WPARAM wParam, 
                             LPARAM lParam) 
{
   HMENU hMenu;

   switch (message)
   {
      case WM_INITDIALOG:  // message: initialize dialog box
         return TRUE;

      case WM_COMMAND:     // message: received a command
         switch (wParam)
         {
            case IDOK:
            {
               char  szName[STR_LEN];  // mailslot name
               char  szTitle[STR_LEN]; // window caption
               int   nErr;             // return code from mailslot API

               // get the contents of the "Mailslot name:" edit control

               GetDlgItemText(hDlg, IDC_MAKEEDIT, (LPSTR)szName, STR_LEN);

               if (strlen(szName) == 0)
               {
                  MessageBox(hDlg, (LPSTR)"Please fill in the edit box.", NULL, MB_OK | MB_ICONSTOP);
                  return TRUE;
               }

               // create the mailslot using the user-specified name

               nErr = DosMakeMailslot(szName,      // name of mailslot
                                      MAX_LEN,     // max message size (bytes)
                                      0,           // mailslot size
                                      &hMailslot); // recieves mailslot handle

               if (nErr != NERR_Success)
                  ErrorMB(hDlg, nErr);
               else
               {
                  // set the window caption to display the mailslot name

                  wsprintf((LPSTR)szTitle, "Mailslot name: %s", (LPSTR)szName);
                  SetWindowText(GetParent(hDlg), (LPSTR)szTitle);

                  // enable and disable the appropriate menuitems

                  hMenu = GetMenu(GetParent(hDlg));
                  EnableMenuItem(hMenu,
                                 IDM_DELETEMAILSLOT,
                                 MF_BYCOMMAND | MF_ENABLED);
                  EnableMenuItem(hMenu,
                                 IDM_READMAILSLOT,
                                 MF_BYCOMMAND | MF_ENABLED);
                  EnableMenuItem(hMenu,
                                 IDM_PEEKMAILSLOT,
                                 MF_BYCOMMAND | MF_ENABLED);
                  EnableMenuItem(hMenu,
                                 IDM_MAILSLOTINFO,
                                 MF_BYCOMMAND | MF_ENABLED);
                  EnableMenuItem(hMenu,
                                 IDM_MAKEMAILSLOT,
                                 MF_BYCOMMAND | MF_GRAYED);

                  // end the dialog box

                  EndDialog(hDlg, TRUE);
                  return TRUE;
               }
            }
            break;

            case IDCANCEL:
            {
               EndDialog(hDlg, FALSE);  // Exits the dialog box
               return TRUE;
            }
         }
         break;
   }

   return FALSE;           // Didn't process a message
}

//****************************************************************************
// Function: WriteDlgProc()
//
// Purpose:  Processes messages for "DosWriteMailslot" dialog box
//
// Parameters:
//    HWND   hDlg
//    UINT   message
//    WPARAM wParam
//    LPARAM lParam
//
// Returns:
//    BOOL
//
//****************************************************************************
BOOL FAR PASCAL WriteDlgProc (HWND   hDlg,
                              UINT   message,
                              WPARAM wParam, 
                              LPARAM lParam) 
{
   switch (message)
   {
      case WM_INITDIALOG:  // message: initialize dialog box
         return TRUE;

      case WM_COMMAND:     // message: received a command
         switch (wParam)
         {
            case IDOK:
            {
               char  szName[STR_LEN];  // mailslot or computer\mailslot name
               char  szMsg[MAX_LEN];   // message to write to the mailslot
               int   nErr;             // return code from mailslot API

               // get the contents of the "Mailslot name:" and "Message:"
               // edit controls

               GetDlgItemText(hDlg, IDC_NAMEEDIT, (LPSTR)szName, STR_LEN);
               GetDlgItemText(hDlg, IDC_MSGEDIT, (LPSTR)szMsg, MAX_LEN);

               if (strlen(szName) == 0 || strlen(szMsg) == 0)
               {
                  MessageBox(hDlg, (LPSTR)"Please fill in both edit boxes.", NULL, MB_OK | MB_ICONSTOP);
                  return TRUE;
               }

               // write the specified message to the specified mailslot

               nErr = DosWriteMailslot(szName,          // mailslot name
                                       szMsg,           // pointer to data to write
                                       strlen(szMsg)+1, // size of data
                                       9,               // priority
                                       2,               // class
                                       0L);             // timeout

               if (nErr != NERR_Success)
                  ErrorMB(hDlg, nErr);
               else
               {
                  // succeeded, so end the dialog box

                  EndDialog(hDlg, TRUE);
                  return TRUE;
               }
            }
            break;

            case IDCANCEL:
            {
               EndDialog(hDlg, FALSE);  // Exits the dialog box
               return TRUE;
            }
         }
         break;
   }

   return FALSE;           // Didn't process a message
}

//****************************************************************************
// Function: ErrorMB()
//
// Purpose:  Displays a message box depending on the given error code.
//
// Parameters:
//    HWND   hWnd
//    UINT   nErr
//
// Returns:
//    nothing
//
//****************************************************************************
void ErrorMB (HWND hWnd, UINT nErr)
{
   switch (nErr)
   {
      case ERROR_ALREADY_EXISTS:
         MessageBox(hWnd, (LPSTR)"Mailslot already exists!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_BAD_NETPATH:
         MessageBox(hWnd, (LPSTR)"Network path was not found!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_BROKEN_PIPE:
         MessageBox(hWnd, (LPSTR)"Pipe is closed or not being read!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_INTERRUPT:
         MessageBox(hWnd, (LPSTR)"Call was interrupted!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_INVALID_HANDLE:
         MessageBox(hWnd, (LPSTR)"Handle specified is invalid!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_INVALID_FUNCTION:
         MessageBox(hWnd,
                    (LPSTR)"Function is invalid such as when the pszName parameter specifies a computer, but the computer is not started!",
                    NULL,
                    MB_OK | MB_ICONSTOP);
         break;

      case ERROR_INVALID_PARAMETER:
         MessageBox(hWnd, (LPSTR)"Parameter specified is invalid!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_NETWORK_BUSY:
         MessageBox(hWnd, (LPSTR)"Network is busy!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_PATH_NOT_FOUND:
         MessageBox(hWnd, (LPSTR)"Path not found!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case ERROR_SEM_TIMEOUT:
         MessageBox(hWnd, (LPSTR)"Time-out occurred!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case NERR_NetNotStarted:
         MessageBox(hWnd, (LPSTR)"Network is not started!", NULL, MB_OK | MB_ICONSTOP);
         break;

      case NERR_Success:
         MessageBox(hWnd, (LPSTR)"Command completed successfully!", (LPSTR)"Success", MB_OK | MB_ICONSTOP);
         break;

      default:
         MessageBox(hWnd, (LPSTR)"Unknown error occurred!", NULL, MB_OK | MB_ICONSTOP);
   }
}
