//****************************************************************************
// File: wndproc.c
//
// Purpose: Main overlapped window's message handler
//
// Functions:
//    MainWndProc() - message handler for main overlapped window
//    DoCommands() - called by MainWndProc to handle all WM_COMMAND messages
//
//****************************************************************************

#include <windows.h>
#include <netcons.h>    // required for API_FUNCTION #def
#include <mailslot.h>   // mailslot API prototypes
#include <neterr.h>     // required for NERR_* errors
#include "global.h"
#include "about.h"
#include "dlgs.h"
#include "wndproc.h"

UINT           hMailslot=NULL;   // handle to mailslot
static HGLOBAL hBuf;        // handle to global memory block for storing mailslot msgs
static LPSTR   lpBuf;       // pointer to global memory for storing mailslot msgs
static BOOL    bRead=FALSE; // flag for tracking if a message has been read yet
static UINT    cbMsgSize, cbNextSize, usNextPriority;

// Local function prototypes
long DoCommands (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

//****************************************************************************
// Function: MainWndProc
//
// Purpose: Message handler for main overlapped window.
//
// Parameters:
//    hWnd    == Handle to _this_ window.
//    message == Message to process.
//    wParam  == WORD parameter -- depends on message
//    lParam  == LONG parameter -- depends on message
//
// Returns: Depends on message.
//
//****************************************************************************
long FAR PASCAL MainWndProc (HWND   hWnd,
                             UINT   message,
                             WPARAM wParam,
                             LPARAM lParam)
{
   switch (message)
   {
      case WM_COMMAND:
         return DoCommands(hWnd, message, wParam, lParam);

      case WM_PAINT:
         {
            PAINTSTRUCT ps;
            RECT        rc;

            BeginPaint(hWnd, &ps);
            if (bRead)
            {
               GetClientRect(hWnd, &rc);
               rc.left   += 10;
               rc.top    += 10;
               rc.right  -= 10;
               rc.bottom -= 10;

               // output the mailslot message
               DrawText(ps.hdc, lpBuf, cbMsgSize-1, &rc, DT_WORDBREAK);
            }
            EndPaint(hWnd, &ps);
         }
         break;

      case WM_CREATE:
         // allocate memory block for storing mailslot messages
         hBuf  = GlobalAlloc(GHND, MAX_LEN);
         lpBuf = (LPSTR)GlobalLock(hBuf);
         break;

      case WM_CLOSE:
         DestroyWindow(hWnd);
         return 0L;

      case WM_DESTROY:
         // free memory block
         GlobalUnlock(hBuf);
         GlobalFree(hBuf);
         PostQuitMessage(0);
         break;

      default:    
         return DefWindowProc(hWnd, message, wParam, lParam);
   }

   return (NULL);
}

//****************************************************************************
// Function: DoCommands
//
// Purpose: Called by MainWndProc() to handle all WM_COMMAND messages.
//
// Parameters:
//    hWnd    == Handle to _this_ window.
//    message == Message to process.
//    wParam  == WORD parameter -- depends on message
//    lParam  == LONG parameter -- depends on message
//
// Returns : Depends on message.
//
//****************************************************************************
long DoCommands (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   switch (wParam)
   {
      case IDM_MAKEMAILSLOT:
         {
            FARPROC lpProcMakeDlg;

            // bring up the "Create a Mailslot" dialog

            lpProcMakeDlg = MakeProcInstance(MakeDlgProc, ghInst);
            DialogBox(ghInst,          // current instance
                      "MakeDlg",       // resource to use
                      hWnd,            // parent handle
                      lpProcMakeDlg);  // MakeDlg() instance address
            FreeProcInstance(lpProcMakeDlg);
         }
         break;

      case IDM_DELETEMAILSLOT:
         {
            int   nErr;              // return code from mailslot API
            char  szTitle[STR_LEN];  // original window caption

            // delete the existing mailslot

            nErr = DosDeleteMailslot(hMailslot);

            if (nErr != NERR_Success)
               ErrorMB(hWnd, nErr);
            else
            {
               HMENU hMenu; // handle to main window's menu

               hMailslot = NULL;

               // restore the original window caption

               LoadString(ghInst, IDS_PROGNAME, szTitle, STR_LEN);
               SetWindowText(hWnd, szTitle);

               // enable and disable the appropriate menuitems

               hMenu = GetMenu(hWnd);
               EnableMenuItem(hMenu,
                              IDM_MAKEMAILSLOT,
                              MF_BYCOMMAND | MF_ENABLED);
               EnableMenuItem(hMenu,
                              IDM_DELETEMAILSLOT,
                              MF_BYCOMMAND | MF_GRAYED);
               EnableMenuItem(hMenu,
                              IDM_READMAILSLOT,
                              MF_BYCOMMAND | MF_GRAYED);
               EnableMenuItem(hMenu,
                              IDM_PEEKMAILSLOT,
                              MF_BYCOMMAND | MF_GRAYED);
               EnableMenuItem(hMenu,
                              IDM_MAILSLOTINFO,
                              MF_BYCOMMAND | MF_GRAYED);
            }
         }
         break;

      case IDM_WRITEMAILSLOT:
         {
            FARPROC lpProcWriteDlg;

            // bring up the "Write to a Mailslot" dialog

            lpProcWriteDlg = MakeProcInstance(WriteDlgProc, ghInst);
            DialogBox(ghInst,          // current instance
                      "WriteDlg",      // resource to use
                      hWnd,            // parent handle
                      lpProcWriteDlg); // WriteDlg() instance address
            FreeProcInstance(lpProcWriteDlg);
         }
         break;

      case IDM_READMAILSLOT:
         {
            int nErr; // return code from mailslot API

            // read the current message in the mailslot

            nErr = DosReadMailslot(hMailslot,      // handle to mailslot
                                   lpBuf,          // pointer to buffer
                                   &cbMsgSize,     // size of message
                                   &cbNextSize,    // size of next message
                                   &usNextPriority,// priority of next message
                                   0L);            // timeout

            if (nErr != NERR_Success)
            {
               // error, no read occurred

               bRead = FALSE;
               ErrorMB(hWnd, nErr);
            }
            else
            {
               // succeeded, so force a redraw to output the message
               // on the main window's client area

               if (cbMsgSize)
               {
                  bRead = TRUE;
                  InvalidateRect(hWnd, NULL, TRUE);
                  UpdateWindow(hWnd);
               }
               else
                  MessageBox(hWnd,
                             (LPSTR)"No more messages in this mailslot!",
                             (LPSTR)"Read Mailslot",
                             MB_OK | MB_ICONSTOP);
            }
         }
         break;

      case IDM_PEEKMAILSLOT:
         {
            HGLOBAL hTmp;  // handle to memory block for storing peeked msg
            LPSTR   lpTmp; // pointer to memory block for storing peeked msg
            int     nErr;  // return code from mailslot API
            UINT    cbSize, cbNextSize, usNextPriority;

            // allocate and lock the temp memory block

            hTmp  = GlobalAlloc(GHND, MAX_LEN);
            lpTmp = (LPSTR)GlobalLock(hTmp);

            // read the highest-priority, most current message in
            // the mailslot without removing it

            nErr = DosPeekMailslot(hMailslot,        // handle to the mailslot
                                   lpTmp,            // buffer to receive message
                                   &cbSize,          // size of message
                                   &cbNextSize,      // size of next message
                                   &usNextPriority); // priority of next message

            if (nErr != NERR_Success)
            {
               // error, no read occurred

               ErrorMB(hWnd, nErr);
            }
            else
            {
               // succeeded, so display message

               if (cbSize)
               {
                  char szPriority[10];

                  wsprintf(szPriority, "Priority: %u", usNextPriority);
                  MessageBox(hWnd, lpTmp, (LPSTR)"Peek", MB_OK);
                  MessageBox(hWnd, (LPSTR)szPriority, (LPSTR)"Peek", MB_OK);
               }
               else
                  MessageBox(hWnd,
                             (LPSTR)"No more messages in this mailslot!",
                             (LPSTR)"Peek at Mailslot",
                             MB_OK | MB_ICONSTOP);
            }

            // clean up

            GlobalUnlock(hTmp);
            GlobalFree(hTmp);
         }
         break;

      case IDM_MAILSLOTINFO:
         {
            UINT cbMessageSize, cbMailslotSize, cbNextSize;
            UINT usNextPriority, cMessages;
            int  nErr; // return code from mailslot API

            // get info about this mailslot

            nErr = DosMailslotInfo(hMailslot,         // handle to the mailslot
                                   &cbMessageSize,    // max size of the messages
                                   &cbMailslotSize,   // size of this mailslot
                                   &cbNextSize,       // size of next message
                                   &usNextPriority,   // priority of next message
                                   &cMessages);       // number of messages in this mailslot

            if (nErr != NERR_Success)
            {
               // error, couldn't get info

               ErrorMB(hWnd, nErr);
            }
            else
            {
               char szInfo[MAX_LEN];

               // succeeded, so display info

               wsprintf(szInfo,
                        "Max msg size: %u\rMailslot size: %u\rSize of next msg: %u\rPriority of next msg: %u\rNumber of msgs in mailslot: %u",
                        cbMessageSize,
                        cbMailslotSize,
                        cbNextSize,
                        usNextPriority,
                        cMessages);

               MessageBox(hWnd,
                          (LPSTR)szInfo,
                          (LPSTR)"Mailslot Info",
                          MB_OK);
            }

         }
         break;

      // Put up the About box.
      case IDM_ABOUT:
      {
         FARPROC lpDlgProc;

         lpDlgProc = MakeProcInstance(AboutDlg, ghInst);
         DialogBox(ghInst,             
                   "AboutBox",         
                   hWnd,               
                   lpDlgProc);         

         FreeProcInstance(lpDlgProc);
         break;
      }

      case IDM_EXIT:
         PostMessage(hWnd, WM_CLOSE, 0, 0L);
         break;

      default:
         return DefWindowProc(hWnd, message, wParam, lParam);
   }

   return (NULL);
}
